// This file is part of OpenCV project.
// It is subject to the license terms in the LICENSE file found in the top-level directory
// of this distribution and at http://opencv.org/license.html

#ifndef OPENCV_GEOMETRY_HPP
#define OPENCV_GEOMETRY_HPP

/**
@defgroup geometry Computational geometry primitives module.
*/

//! @addtogroup geometry
//! @{

#include "opencv2/geometry/2d.hpp"
#include "opencv2/geometry/3d.hpp"

//! @} geometry

#endif
